import os
from typing import Any, Dict, Callable, Union

import numpy
import numpy as np

from sohstationviewer.model.mseed_data.record_reader_helper import Unpacker
from sohstationviewer.model.reftek.from_rt2ms.packet import \
    eh_et_payload_end_in_packet
from sohstationviewer.model.reftek.rt130_experiment.dt_packet import (
    read_dt_packet, DTPacket,
)
from sohstationviewer.model.reftek.rt130_experiment.eh_et_packet import (
    read_eh_et_packet, EHETPacket,
)
from sohstationviewer.model.reftek.rt130_experiment.header import \
    get_rt130_packet_header
from sohstationviewer.model.reftek.rt130_experiment.soh_packets import (
    read_soh_packet, SOHPacket,
)


def packet_reader_placeholder(*args: Any, **kwargs: Any) -> None:
    """
    Placeholder function to be used in place of an RT130 packet reader
    function. This function immediately returns None.
    """
    return None


class RT130ParseError(Exception):
    pass


def read_rt130_file(file_name: str, unpacker: Unpacker):
    # RT130 data looks to be all big-endian (logpeek assumes this, and it has
    # been working pretty well), so we don't have to do any endianness check.

    # if os.path.getsize(file_name) % 1024:
    #     warnings.warn('The size of the data is not a multiple of 1024; it '
    #                   'might be truncated.')

    packets = []

    with open(file_name, 'rb') as rt130_file:
        for i in range(os.path.getsize(file_name) // 1024):
            packet = rt130_file.read(1024)
            packet_header = get_rt130_packet_header(packet, unpacker)

            waveform_handlers: Dict[str, Callable] = {
                'EH': read_eh_et_packet,
                'ET': read_eh_et_packet,
                'DT': read_dt_packet,
            }

            soh_handlers: Dict[str, Callable] = dict.fromkeys(
                ['AD', 'CD', 'DS', 'FD', 'OM', 'SC', 'SH'], read_soh_packet
            )

            packet_handlers = {
                **waveform_handlers, **soh_handlers
            }

            packet_handler = packet_handlers.get(
                packet_header.packet_type, packet_reader_placeholder
            )
            return_val = packet_handler(packet, unpacker)
            if packet_header.packet_type == 'DT':
                packet_type = DTPacket
            elif packet_header.packet_type in ['EH', 'ET']:
                packet_type = EHETPacket
            else:
                packet_type = SOHPacket

            extended_header, data = return_val
            current_packet = packet_type(packet_header, extended_header, data)
            packets.append(current_packet)

    return packets


def convert_packet_to_obspy_format(packet: Union[EHETPacket, DTPacket],
                                   unpacker: Unpacker):
    # We want to convert the packet to a tuple. In order to make it easier to
    # maintain, we first convert the packet to a dictionary. Then, we grab the
    # values of the dictionary as tuple to get the final result.
    converted_packet = {}
    converted_packet['packet_type'] = packet.header.packet_type
    converted_packet['experiment_number'] = packet.header.experiment_number
    # Obspy only stores the last two digits of the year.
    converted_packet['year'] = packet.header.time.year % 100
    converted_packet['unit_id'] = packet.header.unit_id
    converted_packet['time'] = packet.header.time.ns
    converted_packet['byte_count'] = packet.header.byte_count
    converted_packet['packet_sequence'] = packet.header.packet_sequence
    converted_packet['event_number'] = packet.extended_header.event_number
    converted_packet[
        'data_stream_number'] = packet.extended_header.data_stream_number
    converted_packet['channel_number'] = packet.extended_header.channel_number
    converted_packet[
        'number_of_samples'] = packet.extended_header.number_of_samples
    converted_packet['flags'] = packet.extended_header.flags
    converted_packet['data_format'] = packet.extended_header.data_format

    if converted_packet['packet_type'] == 'DT':
        # Obspy stores the data as list of 1-byte integers. We store the
        # data as an arbitrary length integer, so we need to do some
        # conversion. To make converting the resulting tuple to an element
        # of a structured array of type PACKET_FINAL_DTYPE easier, we set
        # the size of the payload to be 4. This only affect data with format
        # 16, and as long as we are careful in self.to_stream, we don't even
        # have to make a special case when decoding (note: this is possible
        # because of a peculiarity of the 2's complement encoding).
        data_size = 4
        format_char = 'B'
        converted_packet['payload'] = numpy.empty(1000, np.uint8)
        packet_data = list(unpacker.unpack(
            f'{data_size}{format_char}',
            packet.data.to_bytes(data_size, 'big', signed=True)
        ))
        converted_packet['payload'][:4] = packet_data
    elif converted_packet['packet_type'] in ['EH', 'ET']:
        eh_et_payload_size = eh_et_payload_end_in_packet - 24
        converted_packet['payload'] = numpy.empty(1000, np.uint8)
        packet_data = numpy.frombuffer(packet.data, np.uint8)
        converted_packet['payload'][:eh_et_payload_size] = packet_data
    else:
        converted_packet['payload'] = numpy.frombuffer(packet.data, np.uint8)
    return tuple(converted_packet.values())
